/**
 * Compiler written in Node.js.
 * Compiles one or multiple Kobold Adventure scenes into a single file.
 */

// Arguments used by the program.
var args = require('minimist')(process.argv.slice(2));
// Framework for pretty printing.
var columnify = require('columnify');
// Framework for colorful output.
var colors = require('colors/safe');
// Internal utils for object debugging.
var util = require('util');
// Path util for constructing file paths.
var path = require('path');
// Filesystem util required for working with files.
var fs = require('fs');

// The working directory of the terminal invoking this script.
var exeDir = path.resolve(process.cwd());
// The file extension of every file denoted in Scene-JSONs.
var sceneFileExtension = ".js";
// Import the SceneMap class from the game.
eval(fs.readFileSync(path.resolve(path.join(exeDir, "res", "js", "SceneMap.js")), 'utf8'));
var sceneMap = new SceneMap();

/**
 * Iterates over the passed array of elements which may have embedded arrays, and applies the function to each element.
 * @param {function} fun The function to apply. Will be passed the current element.
 * @param {array} elems The elements to iterate over.
 */
function recArrayTraverse(fun, elems) {
    for (var i = 0; i < elems.length; i++) {
        var elem = elems[i];
        if (elem && elem.constructor === Array) {
            recArrayTraverse(fun, elem);
        } else {
            fun(elem);
        }
    }
}

/**
 * Logs the passed strings as errors, and then terminates the program.
 * @param {string} strings The string(s) to log. Multiple strings may be passed. Recursive arrays of strings may also be passed.
 */
function error(...strings) {
    recArrayTraverse(elem => console.log(colors.red("ERROR: " + elem)), strings);
    process.exit(1);
}

/**
 * Logs the passed strings as warnings.
 * Does nothing if output is disabled.
 * @param {string} strings The string(s) to log. Multiple strings may be passed. Recursive arrays of strings may also be passed.
 */
function warn(...strings) {
    if (args.q) {
        return;
    }
    recArrayTraverse(elem => console.log(colors.yellow("WARNING: " + elem)), strings);
}

/**
 * Logs the passed strings.
 * Does nothing if output is disabled.
 * @param {string} strings The string(s) to log. Multiple strings may be passed. Recursive arrays of strings may also be passed.
 */
function log(...strings) {
    if (args.q) {
        return;
    }
    recArrayTraverse(elem => console.log(elem), strings);
}

/**
 * Logs the passed strings as debug messages.
 * Does nothing if debugging or output in general is disabled.
 * @param {string} strings The string(s) to log. Multiple strings may be passed. Recursive arrays of strings may also be passed.
 */
function debug(...strings) {
    if (args.q || !args.debug) {
        return;
    }
    recArrayTraverse(elem => console.log(colors.green("DEBUG: " + elem)), strings);
}

/**
 * Prints the help dialogue.
 */
function printHelp() {
    var argumentHelp = {
        "targets": "Everything without a flag is a scene to be compiled by aSceneCompiler. No filenames means aSceneCompiler will do nothing, unless the --all flag was passed.",
        "--all": "Compiles every scene into a single file. Overrides targets.",
        "--ind": "Compiles every scene individually. Overrides output options.",
        "--indout folder": "Specifies a single folder to output all invidually compiled scenes to. No effect without --ind.",
        "--ext extension": "Will overwrite the default .js extension every file in a scene's descriptor .json is assumed to have. Should start iwht a period.",
        "--out filename": "Will output results to the specified file (creating it if need be). Without --out, the output will be printed in the console.",
        "-h": "Displays this help message. Does not run aSceneCompiler.",
        "-q": "Runs aSceneCompiler silently (with no output).",
        "--debug": "Displays debug information while running."
    };
    log("aSceneCompiler is a compiler for Kobold Adventure scenes.");
    log("It reads descriptor json files and concatenates all files referenced within.");
    log("This tool MUST be ran from the root of the Kobold Adventure file structure.");
    log(columnify(argumentHelp, {columns: ['FLAG', 'EXPLANATION'], config: {FLAG: {minWidth: 20}, EXPLANATION: {maxWidth: process.stdout.columns - 40}}}));
}

/**
 * Checks if the passed parameter value is a string.
 * If it's not, the program exits with an error.
 * @param {string} toCheck The variable to check.
 * @param {string} flag Which flag is being checked. Produces a more informative error message.
 */
function stringOrExit(toCheck, flag) {
    if (typeof toCheck !== "string") {
        error("", `Parameter ${flag} must be a string value.`, `Example: ${flag} test`, "");
    }
}

/**
 * Gathers and returns all of the initial scene targets for aSceneCompiler by inspecting command line parameters.
 * @returns {array} An array of scene names to target.
 */
function buildInitialTargets() {
    debug("", "Building initial targets.");
    var input = args._;
    var output = [];

    if (args.all) {
        debug("Compiling all scenes.");
        var allScenes = sceneMap.getAll();
        for (var i = 0; i < allScenes.length; i++) {
            output.push(allScenes[i]);
        }
    } else if (input && input.length) {
        debug("Found targets in arguments.");
        for (var i = 0; i < input.length; i++) {
            output.push(input[i]);
        }
    } else {
        error("", "No targets specified.", "Add either --all or the names of scenes you want to compile to the command line parameters.", "");
    }

    debug("Initial targets:", output, "");
    return output;
}

/**
 * Returns true if the target is an existing file or folder, false otherwise.
 * @param {string} file A path to a file or folder.
 * @returns {boolean} True if the target exists, false otherwise.
 */
function exists(file) {
    return fs.existsSync(file);
}

/**
 * Exits the program with an error message if the passed file does not exist.
 * @param {string} file A path to a file or folder.
 */
function existsOrExit(file) {
    if (!exists(file)) {
        error("", "File not found:", file, "");
    }
}

/**
 * Exits the program with an error message if the passed scene does not exist.
 * @param {string} scene A scene name.
 */
function sceneExistsOrExit(scene) {
    if (!sceneMap.exists(scene)) {
        error("", "Scene not found in SceneMap:", scene, "");
    }
}

/**
 * Gets filestats for the specified file or folder.
 * @param {string} file The path to the file or folder to get stats for.
 * @returns {fs.Stats} The file's stats.
 */
function fileStats(file) {
    return fs.lstatSync(file);
}

/**
 * Returns true if the file associated with these stats is a folder.
 * @param {fs.Stats} stats The file's stats.
 * @returns {boolean} True if the file is a folder, false otherwise.
 */
function isFolder(stats) {
    return stats.isDirectory();
}

/**
 * Compiles a single file, if it has not yet been encountered.
 * Prints a warning if the file has already been compiled.
 * @param {string} target A path to the file to compile.
 * @returns {string} The compiled file. The empty string if the file has already been compiled.
 */
function compileFile(target) {
    var output = "";
    if (compile.visited.includes(target)) {
        warn("File already compiled: " + target);
        return output;
    }
    compile.visited.push(target);
    log("Compiling " + target);
    existsOrExit(target);
    var contents = fs.readFileSync(target, 'utf8');
    var relative = target.replace(exeDir, "").substr(1);
    output += "// " + relative + "\n\n";
    output += contents.trim();
    output += "\n\n// END " + relative + "\n\n";
    return output;
}

/**
 * Loads and returns the array contained within the JSON file for the specified scene.
 * Throws exits with an error if the JSON file does not exists.
 * @param {string} scene The string for which to load the JSON file.
 * @returns {array} The array contained within the JSON file.
 */
function loadSceneJson(scene) {
    var jsonPathRelative = sceneMap.constructJsonPath(scene);
    var jsonPath = path.resolve(path.join(exeDir, jsonPathRelative));
    existsOrExit(jsonPath);
    debug("Reading " + jsonPathRelative);
    var sceneJson = fs.readFileSync(jsonPath, 'utf8');
    try {
        sceneJson = JSON.parse(sceneJson);
    } catch (e) {
        error("", "Failed to parse build file " + jsonPath, "Make sure the JSON file is well-structured.", "");
    }
    debug("Parsed: " + util.inspect(sceneJson));
    return sceneJson;
}

/**
 * Compiles a single scene.
 * Returns the compiled file as a string.
 * Exits the program with an error if the scene does not exist or if the scene JSON is malformed.
 * @param {string} scene The scene to compile.
 * @returns {string} The compiled scene.
 */
function compileScene(scene) {
    var output = "";
    output += "\n\n// SCENE " + scene + "\n\n";
    log("Compiling " + scene);
    sceneExistsOrExit(scene);
    var filenames = loadSceneJson(scene);
    if (!filenames || filenames.constructor !== Array) {
        error("", "Scene JSON was not valid.", "The scene JSON must consist of a single array with values therein.", "");
    }
    var basePath = path.join(exeDir, sceneMap.constructFolderPath(scene));
    for (var i = 0; i < filenames.length; i++) {
        output += compileFile(path.join(basePath, filenames[i] + sceneFileExtension));
    }
    output += "\n\n// END SCENE " + scene + "\n\n";
    return output;
}

/**
 * Compiles all the specified scenes individually.
 * Returns the compiled scenes as an object, with the keys being the scene names.
 * initCompile() should always be called before the initial call to this function.
 * @param {array} scenes All the scenes to compile. These must exist in the scene map.
 * @returns {string} The compiled scenes as an object, with the keys being the scene names.
 */
function compileIndividually(scenes) {
    var output = {};
    for (var i = 0; i < scenes.length; i++) {
        output[scenes[i]] = compileScene(scenes[i]);
    }
    return output;
}

/**
 * Compiles all the specified scenes.
 * Returns the compiled string.
 * initCompile() should always be called before the initial call to this function.
 * @param {array} scenes All the scenes to compile. These must exist in the scene map.
 * @returns {string} The compiled file as a single large string.
 */
function compile(scenes) {
    var output = "";
    for (var i = 0; i < scenes.length; i++) {
        output += compileScene(scenes[i]);
    }
    return output;
}

/**
 * Initializes things needed for compilation.
 * Should always be called before the initial call to compile.
 */
function initCompile() {
    compile.visited = [];
    log("Compiler initialized.");
}

/**
 * Outputs compiled scenes to multiple files in a single folder.
 * @param {object} compiled The compiled scenes, as an object mapping scene names to strings.
 * @param {string} folder The folder to output to.
 */
function individualOutputSingleFolder(compiled, folder) {
    var absoluteFolder = path.resolve(folder);
    log("Exporting individual scenes to " + absoluteFolder);
    for (var scene in compiled) {
        var fileName = path.join(absoluteFolder, scene + "_compiled.js");
        log("Exporting to " + fileName);
        fs.writeFileSync(fileName, compiled[scene]);
    }
    log("Output files successfully created.");
}

/**
 * Outputs compiled scenes to multiple files in multiple folders.
 * @param {object} compiled The compiled scenes, as an object mapping scene names to strings.
 */
function individualOutputMultiFolder(compiled) {
    for (var scene in compiled) {
        var sceneRoot = path.resolve(sceneMap.constructFolderPath(scene));
        var fileName = path.join(sceneRoot, scene + "_compiled.js");
        log("Exporting to " + fileName);
        fs.writeFileSync(fileName, compiled[scene]);
    }
    log("Output files successfully created.");
}

/**
 * Outputs compiled scenes to multiple files.
 * @param {object} compiled The compiled scenes, as an object mapping scene names to strings.
 */
function individualOutput(compiled) {
    if (args.indout) {
        individualOutputSingleFolder(compiled, args.indout);
    } else {
        individualOutputMultiFolder(compiled);
    }
}

/**
 * Outputs compiled scenes to a single file.
 * @param {string} compiled The compiled scenes, as a string.
 */
function fileOutput(compiled) {
    var outputFileName = path.resolve(args.out);
    log("Exporting to " + outputFileName);
    fs.writeFileSync(outputFileName, compiled);
    log("Output file successfully created.");
}

/**
 * Outputs compiled scenes to the console.
 * @param {string} compiled The compiled scenes, as a string.
 */
function consoleOutput(compiled) {
    log(compiled);
}

if (args.h) {
    printHelp();
    process.exit(0);
}

log(colors.cyan.underline("aSceneCompiler"));
debug("", "----------------------", "Debug logging started.", "----------------------", "");
debug("Arguments:", util.inspect(args), "");

if (args.ext) {
    stringOrExit(args.ext, "--ext");
    sceneFileExtension = args.ext;
}

if (args.out) {
    stringOrExit(args.out, "--out");
}

if (args.indout) {
    stringOrExit(args.indout, "--indout");
    existsOrExit(args.indout);
    if (!isFolder(fileStats(args.indout))) {
        error("", "Invalid output target: " + args.indout, "Independent compilation output target must be a folder.", "");
    }
}

initCompile();
console.time("Duration");
log("Commencing compilation.");
var compiled = args.ind ? compileIndividually(buildInitialTargets()) : compile(buildInitialTargets());
log("Compilation complete.");
console.timeEnd("Duration");

if (args.ind) {
    individualOutput(compiled);
} else if (args.out) {
    fileOutput(compiled);
} else {
    consoleOutput(compiled);
}

