/**
 * Class that represents a single theme.
 * By default, all values will be akin to the main theme of Kobold Adventure.
 * @param {object} colors An object containing the colors to be used in this theme.
 *                        Every color must be either an {r, g, b} object, a ThemeColor object or a hexadecimal string.
 *                        Every color is optional. If a color is not specified, the default from the main theme will be used.
 *                        Available options are:
 *                          -background
 *                          -main
 *                          -detail
 *                          -text
 *                          -rare
 *                          -danger
 *                          -sneaky
 *                          -blood
 *                          -sanity
 *                          -energy
 *                          -exertion
 * @param {string} additionalRules Any additional custom CSS rules to add to this theme in the form of a string.
 */
function Theme(colors, additionalRules) {
    colors = colors || {};
    for (var option in colors) {
        if (!colors[option]) {
            continue;
        } else if (colors[option].constructor === ThemeColor) {
            continue;
        } else if (typeof colors[option] === "string") {
            colors[option] = new ThemeColor(colors[option]);
        } else {
            colors[option] = new ThemeColor(colors[option]);
        }
    }

    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";
    this.background = colors.background || new ThemeColor("#191919");
    this.main = colors.main || new ThemeColor("#005AA1");
    this.detail = colors.detail || new ThemeColor("#003C6C");
    this.text = colors.text || new ThemeColor("#FFFFFF");
    this.rare = colors.rare || new ThemeColor("#FF6EC7");
    this.danger = colors.danger || new ThemeColor("#BE2625");
    this.sneaky = colors.sneaky || new ThemeColor("#C0C0C0");
    this.blood = colors.blood || new ThemeColor("#AF111C");
    this.sanity = colors.sanity || new ThemeColor("#4B0082");
    this.energy = colors.energy || new ThemeColor("#FFCC11");
    this.exertion = colors.exertion || new ThemeColor("#FF6EC7");

    this.extraRules = additionalRules || "";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Loads this object from a type-stripped version thereof.
     * @param {object} theme Type-stripped Theme object.
     */
    this.load = function (theme) {
        var defaults = new Theme();
        if (!theme) {
            that.load(defaults);
            return;
        }
        that.background = theme.background ? new ThemeColor(theme.background) : defaults.background;
        that.main = theme.main ? new ThemeColor(theme.main) : defaults.main;
        that.detail = theme.detail ? new ThemeColor(theme.detail) : defaults.detail;
        that.text = theme.text ? new ThemeColor(theme.text) : defaults.text;
        that.rare = theme.rare ? new ThemeColor(theme.rare) : defaults.rare;
        that.danger = theme.danger ? new ThemeColor(theme.danger) : defaults.danger;
        that.sneaky = theme.sneaky ? new ThemeColor(theme.sneaky) : defaults.sneaky;
        that.blood = theme.blood ? new ThemeColor(theme.blood) : defaults.blood;
        that.sanity = theme.sanity ? new ThemeColor(theme.sanity) : defaults.sanity;
        that.energy = theme.energy ? new ThemeColor(theme.energy) : defaults.energy;
        that.exertion = theme.exertion ? new ThemeColor(theme.exertion) : defaults.exertion;
        that.extraRules = theme.extraRules ? theme.extraRules : defaults.extraRules;
    };
}
