/**
 * This class acts as a dedicated controller for the home part of the Forest scene.
 * @param {string} sceneId The id of the scene.
 */
function ForestHomeController(sceneId) {
    // PRIVATE MEMBERS \\       var private = "test";
    var id = sceneId;
    var pathChoiceId = "homePathChoice";
    var pathStatusChangeId = "homePathStatus";
    var pathWeaponId = "homePathWeapon";
    var hideContinueId = "homeHideContinue";
    var hideStatusChangeId = "homeHideStatus";
    var hideWeaponId = "homeHideWeapon";
    var ignoreContinueId = "homeIgnoreContinue";
    var ignoreStatusChangeId = "homeIgnoreStatus";

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Makes the path choice.
     * @param {ForestHomePathChoiceEnum} choice The choice that was made.
     */
    this.makePathChoice = function (choice) {
        var history = global.kobold.history.get(id);
        history[pathChoiceId] = choice;
        global.kobold.history.set(id, history);
    };

    /**
     * Returns whether or not the path choice has been made.
     * @returns {boolean} Whether or not the path choice has been made.
     */
    this.hasMadePathChoice = function () {
        return global.kobold.history.has(id, pathChoiceId);
    };

    /**
     * Fetches the path choice from history and returns it.
     * May return undefined if the choice has not yet been made.
     * @returns {ForestHomePathChoiceEnum} The choice that was made.
     */
    this.getPathChoice = function () {
        var history = global.kobold.history.get(id);
        return history[pathChoiceId];
    };

    /**
     * Reduces the kobold's sanity if required.
     * @returns {string} HTML describing the change.
     */
    this.pathStatusChange = function () {
        if (!global.kobold.history.has(id, pathStatusChangeId)) {
            global.kobold.status.mental.decrease(5, "Disconcerting scream");
            var history = global.kobold.history.get(id);
            history[pathStatusChangeId] = true;
            global.kobold.history.set(id, history);
        }
        return global.ui.change.buildNotification([
            new Change("Disconcerted", "Sanity", -5)
        ]);
    };

    /**
     * Stores (if required) and retrieves the kobold's weapon at the time of traversing the path.
     * @returns {Gear} The weapon.
     */
    this.pathWeapon = function () {
        if (!global.kobold.history.has(id, pathWeaponId)) {
            var history = global.kobold.history.get(id);
            if (global.kobold.inventory.equipment.isWearing(SlotEnum.WEAPON)) {
                history[pathWeaponId] = global.kobold.inventory.equipment.weapon;
            } else {
                history[pathWeaponId] = null;
            }
            global.kobold.history.set(id, history);
        }
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.get(id)[hideWeaponId]);
    };

    /**
     * Signals that the hide section of the scene has continued.
     */
    this.hideContinue = function () {
        var history = global.kobold.history.get(id);
        history[hideContinueId] = true;
        global.kobold.history.set(id, history);
    };

    /**
     * Returns whether or not the hide section of this part of the scene has continued.
     * @returns {boolean} Whether or not the hide section of this part of the scene has continued.
     */
    this.hideHasContinued = function () {
        return global.kobold.history.has(id, hideContinueId);
    };

    /**
     * Stores (if required) and retrieves the kobold's weapon at the time of hiding in the forest.
     * @returns {Gear} The weapon.
     */
    this.hideWeapon = function () {
        if (!global.kobold.history.has(id, hideWeaponId)) {
            var history = global.kobold.history.get(id);
            if (global.kobold.inventory.equipment.isWearing(SlotEnum.WEAPON)) {
                history[hideWeaponId] = global.kobold.inventory.equipment.weapon;
            } else {
                history[hideWeaponId] = null;
            }
            global.kobold.history.set(id, history);
        }
        return global.kobold.inventory.equipment.buildGearFromHistory(global.kobold.history.get(id)[hideWeaponId]);
    };

    /**
     * Reduces the kobold's sanity if required.
     * @returns {string} HTML describing the change.
     */
    this.hideStatusChange = function () {
        if (!global.kobold.history.has(id, hideStatusChangeId)) {
            global.kobold.status.mental.decrease(10, "Spooky noises");
            var history = global.kobold.history.get(id);
            history[hideStatusChangeId] = true;
            global.kobold.history.set(id, history);
        }
        return global.ui.change.buildNotification([
            new Change("Spooked", "Sanity", -10)
        ]);
    };

    /**
     * Signals that the ignore section of the scene has continued.
     */
    this.ignoreContinue = function () {
        global.kobold.history.doIfNotSet(id, ignoreContinueId, true, function () {
            global.kobold.history.shared.home.status = HomeStatusEnum.SAVED;
        });
    };

    /**
     * Returns whether or not the ignore section of this part of the scene has continued.
     * @returns {boolean} Whether or not the ignore section of this part of the scene has continued.
     */
    this.ignoreHasContinued = function () {
        return global.kobold.history.has(id, ignoreContinueId);
    };

    /**
     * Reduces the kobold's sanity if required.
     * @returns {string} HTML describing the change.
     */
    this.ignoreStatusChange = function () {
        if (!global.kobold.history.has(id, ignoreStatusChangeId)) {
            global.kobold.status.mental.decrease(20, "Crippling guilt");
            var history = global.kobold.history.get(id);
            history[ignoreStatusChangeId] = true;
            global.kobold.history.set(id, history);
        }
        return global.ui.change.buildNotification([
            new Change("Guilt", "Sanity", -20)
        ]);
    };
}
